"""Functions required by the OER C code generator

"""

from .utils import ENCODER_ABORT
from .utils import DECODER_ABORT

ENUMERATED_VALUE_LENGTH = '''
static uint8_t enumerated_value_length(int32_t value)
{
    uint8_t length;

    if ((value >=0) && (value < 128)) {
        length = 0;
    } else if ((value >= -128) && (value < 128)) {
        length = 1;
    } else if ((value >= -32768) && (value < 32768)) {
        length = 2;
    } else if ((value >= -8388608) && (value < 8388608)) {
        length = 3;
    } else {
        length = 4;
    }

    return length;
}\
'''

LENGTH_DETERMINANT_LENGTH = '''
static uint32_t length_determinant_length(uint32_t value)
{
    uint32_t length;

    if (value < 128u) {
        length = 1;
    } else if (value < 256u) {
        length = 2;
    } else if (value < 65536u) {
        length = 3;
    } else if (value < 16777216u) {
        length = 4;
    } else {
        length = 5;
    }

    return (length);
}\
'''

MINIMUM_UINT_LENGTH = '''
static uint8_t minimum_uint_length(uint32_t value)
{
    uint8_t length;

    if (value < 256u) {
        length = 1;
    } else if (value < 65536u) {
        length = 2;
    } else if (value < 16777216u) {
        length = 3;
    } else {
        length = 4;
    }

    return (length);
}\
'''

ENCODER_INIT = '''\
static void encoder_init(struct encoder_t *self_p,
                         uint8_t *buf_p,
                         size_t size)
{
    self_p->buf_p = buf_p;
    self_p->size = (ssize_t)size;
    self_p->pos = 0;
}\
'''

ENCODER_GET_RESULT = '''
static ssize_t encoder_get_result(const struct encoder_t *self_p)
{
    return (self_p->pos);
}\
'''

ENCODER_ALLOC = '''
static ssize_t encoder_alloc(struct encoder_t *self_p,
                             size_t size)
{
    ssize_t pos;

    if ((self_p->pos + (ssize_t)size) <= self_p->size) {
        pos = self_p->pos;
        self_p->pos += (ssize_t)size;
    } else {
        pos = -ENOMEM;
        encoder_abort(self_p, ENOMEM);
    }

    return (pos);
}\
'''

ENCODER_APPEND_BYTES = '''
static void encoder_append_bytes(struct encoder_t *self_p,
                                 const uint8_t *buf_p,
                                 size_t size)
{
    ssize_t pos;

    pos = encoder_alloc(self_p, size);

    if (pos < 0) {
        return;
    }

    (void)memcpy(&self_p->buf_p[pos], buf_p, size);
}\
'''

ENCODER_APPEND_UINT8 = '''
static void encoder_append_uint8(struct encoder_t *self_p,
                                 uint8_t value)
{
    encoder_append_bytes(self_p, &value, sizeof(value));
}\
'''

ENCODER_APPEND_UINT16 = '''
static void encoder_append_uint16(struct encoder_t *self_p,
                                  uint16_t value)
{
    uint8_t buf[2];

    buf[0] = (uint8_t)(value >> 8);
    buf[1] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_UINT32 = '''
static void encoder_append_uint32(struct encoder_t *self_p,
                                  uint32_t value)
{
    uint8_t buf[4];

    buf[0] = (uint8_t)(value >> 24);
    buf[1] = (uint8_t)(value >> 16);
    buf[2] = (uint8_t)(value >> 8);
    buf[3] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_UINT64 = '''
static void encoder_append_uint64(struct encoder_t *self_p,
                                  uint64_t value)
{
    uint8_t buf[8];

    buf[0] = (uint8_t)(value >> 56);
    buf[1] = (uint8_t)(value >> 48);
    buf[2] = (uint8_t)(value >> 40);
    buf[3] = (uint8_t)(value >> 32);
    buf[4] = (uint8_t)(value >> 24);
    buf[5] = (uint8_t)(value >> 16);
    buf[6] = (uint8_t)(value >> 8);
    buf[7] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_INT8 = '''
static void encoder_append_int8(struct encoder_t *self_p,
                                int8_t value)
{
    encoder_append_uint8(self_p, (uint8_t)value);
}\
'''

ENCODER_APPEND_INT16 = '''
static void encoder_append_int16(struct encoder_t *self_p,
                                 int16_t value)
{
    encoder_append_uint16(self_p, (uint16_t)value);
}\
'''

ENCODER_APPEND_INT32 = '''
static void encoder_append_int32(struct encoder_t *self_p,
                                 int32_t value)
{
    encoder_append_uint32(self_p, (uint32_t)value);
}\
'''

ENCODER_APPEND_INT64 = '''
static void encoder_append_int64(struct encoder_t *self_p,
                                 int64_t value)
{
    encoder_append_uint64(self_p, (uint64_t)value);
}\
'''

ENCODER_APPEND_UINT = '''
static void encoder_append_uint(struct encoder_t *self_p,
                                uint32_t value,
                                uint8_t number_of_bytes)
{
    switch (number_of_bytes) {

    case 1:
        encoder_append_uint8(self_p, (uint8_t)value);
        break;

    case 2:
        encoder_append_uint16(self_p, (uint16_t)value);
        break;

    case 3:
        encoder_append_uint8(self_p, (uint8_t)(value >> 16));
        encoder_append_uint16(self_p, (uint16_t)value);
        break;

    default:
        encoder_append_uint32(self_p, value);
        break;
    }
}\
'''

ENCODER_APPEND_LONG_UINT = '''
static void encoder_append_long_uint(struct encoder_t *self_p,
                                     uint64_t value,
                                     uint8_t number_of_bytes)
{
    const uint8_t *value_p = (const uint8_t*)&value;
    uint8_t buf[8];
    for(uint32_t byte = 0; byte < number_of_bytes; byte++) {
        buf[number_of_bytes - byte - 1] = *value_p++;
    }
    encoder_append_bytes(self_p, buf, number_of_bytes);
}\
'''

ENCODER_APPEND_INT = '''
static void encoder_append_int(struct encoder_t *self_p,
                               int32_t value,
                               uint8_t number_of_bytes)
{
    switch (number_of_bytes) {

    case 1:
        encoder_append_int8(self_p, (int8_t)value);
        break;

    case 2:
        encoder_append_int16(self_p, (int16_t)value);
        break;

    case 3:
        encoder_append_uint8(self_p, (uint8_t)((uint32_t)value >> 16));
        encoder_append_int16(self_p, (int16_t)value);
        break;

    default:
        encoder_append_int32(self_p, value);
        break;
    }
}\
'''

ENCODER_APPEND_FLOAT = '''
static void encoder_append_float(struct encoder_t *self_p,
                                 float value)
{
    uint32_t i32;

    (void)memcpy(&i32, &value, sizeof(i32));

    encoder_append_uint32(self_p, i32);
}\
'''

ENCODER_APPEND_DOUBLE = '''
static void encoder_append_double(struct encoder_t *self_p,
                                  double value)
{
    uint64_t i64;

    (void)memcpy(&i64, &value, sizeof(i64));

    encoder_append_uint64(self_p, i64);
}\
'''

ENCODER_APPEND_BOOL = '''
static void encoder_append_bool(struct encoder_t *self_p, bool value)
{
    encoder_append_uint8(self_p, value ? 255u : 0u);
}\
'''

ENCODER_APPEND_LENGTH_DETERMINANT = '''
static void encoder_append_length_determinant(struct encoder_t *self_p,
                                              uint32_t length)
{
    if (length < 128u) {
        encoder_append_int8(self_p, (int8_t)length);
    } else if (length < 256u) {
        encoder_append_uint8(self_p, 0x81u);
        encoder_append_uint8(self_p, (uint8_t)length);
    } else if (length < 65536u) {
        encoder_append_uint8(self_p, 0x82u);
        encoder_append_uint16(self_p, (uint16_t)length);
    } else if (length < 16777216u) {
        encoder_append_uint32(self_p, length | (0x83u << 24u));
    } else {
        encoder_append_uint8(self_p, 0x84u);
        encoder_append_uint32(self_p, length);
    }
}\
'''

DECODER_INIT = '''
static void decoder_init(struct decoder_t *self_p,
                         const uint8_t *buf_p,
                         size_t size)
{
    self_p->buf_p = buf_p;
    self_p->size = (ssize_t)size;
    self_p->pos = 0;
}\
'''

DECODER_GET_RESULT = '''
static ssize_t decoder_get_result(const struct decoder_t *self_p)
{
    return (self_p->pos);
}\
'''

DECODER_FREE = '''
static ssize_t decoder_free(struct decoder_t *self_p,
                            size_t size)
{
    ssize_t pos;

    if ((self_p->pos + (ssize_t)size) <= self_p->size) {
        pos = self_p->pos;
        self_p->pos += (ssize_t)size;
    } else {
        pos = -EOUTOFDATA;
        decoder_abort(self_p, EOUTOFDATA);
    }

    return (pos);
}\
'''

DECODER_READ_BYTES = '''
static void decoder_read_bytes(struct decoder_t *self_p,
                               uint8_t *buf_p,
                               size_t size)
{
    ssize_t pos;

    pos = decoder_free(self_p, size);

    if (pos >= 0) {
        (void)memcpy(buf_p, &self_p->buf_p[pos], size);
    } else {
        (void)memset(buf_p, 0, size);
    }
}\
'''

DECODER_READ_UINT8 = '''
static uint8_t decoder_read_uint8(struct decoder_t *self_p)
{
    uint8_t value;

    decoder_read_bytes(self_p, &value, sizeof(value));

    return (value);
}\
'''

DECODER_READ_UINT16 = '''
static uint16_t decoder_read_uint16(struct decoder_t *self_p)
{
    uint8_t buf[2];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (uint16_t)(((uint16_t)buf[0] << 8) | (uint16_t)buf[1]);
}\
'''

DECODER_READ_UINT32 = '''
static uint32_t decoder_read_uint32(struct decoder_t *self_p)
{
    uint8_t buf[4];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (((uint32_t)buf[0] << 24)
            | ((uint32_t)buf[1] << 16)
            | ((uint32_t)buf[2] << 8)
            | (uint32_t)buf[3]);
}\
'''

DECODER_READ_UINT64 = '''
static uint64_t decoder_read_uint64(struct decoder_t *self_p)
{
    uint8_t buf[8];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (((uint64_t)buf[0] << 56)
            | ((uint64_t)buf[1] << 48)
            | ((uint64_t)buf[2] << 40)
            | ((uint64_t)buf[3] << 32)
            | ((uint64_t)buf[4] << 24)
            | ((uint64_t)buf[5] << 16)
            | ((uint64_t)buf[6] << 8)
            | (uint64_t)buf[7]);
}\
'''

DECODER_READ_INT8 = '''
static int8_t decoder_read_int8(struct decoder_t *self_p)
{
    return ((int8_t)decoder_read_uint8(self_p));
}\
'''

DECODER_READ_INT16 = '''
static int16_t decoder_read_int16(struct decoder_t *self_p)
{
    return ((int16_t)decoder_read_uint16(self_p));
}\
'''

DECODER_READ_INT32 = '''
static int32_t decoder_read_int32(struct decoder_t *self_p)
{
    return ((int32_t)decoder_read_uint32(self_p));
}\
'''

DECODER_READ_INT64 = '''
static int64_t decoder_read_int64(struct decoder_t *self_p)
{
    return ((int64_t)decoder_read_uint64(self_p));
}\
'''

DECODER_READ_UINT = '''
static uint32_t decoder_read_uint(struct decoder_t *self_p,
                                  uint8_t number_of_bytes)
{
    uint32_t value;

    switch (number_of_bytes) {

    case 1:
        value = decoder_read_uint8(self_p);
        break;

    case 2:
        value = decoder_read_uint16(self_p);
        break;

    case 3:
        value = ((uint32_t)decoder_read_uint8(self_p) << 16u);
        value |= decoder_read_uint16(self_p);
        break;

    case 4:
        value = decoder_read_uint32(self_p);
        break;

    default:
        value = 0xffffffffu;
        break;
    }

    return (value);
}\
'''

DECODER_READ_LONG_UINT = '''
static uint64_t decoder_read_long_uint(struct decoder_t *self_p,
                                       uint8_t number_of_bytes)
{
    uint64_t value = 0;

    for(uint8_t byte = 0; byte < number_of_bytes; byte++) {
        value = decoder_read_uint8(self_p) | (value << 8);
    }

    return (value);
}\
'''

DECODER_READ_INT = '''
static int32_t decoder_read_int(struct decoder_t *self_p,
                                uint8_t number_of_bytes)
{
    int32_t value;
    uint32_t tmp;

    switch (number_of_bytes) {

    case 1:
        value = decoder_read_int8(self_p);
        break;

    case 2:
        value = decoder_read_int16(self_p);
        break;

    case 3:
        tmp = ((uint32_t)decoder_read_uint8(self_p) << 16u);
        tmp |= decoder_read_uint16(self_p);
        if((tmp & 0x800000u) == 0x800000u) {
            tmp += 0xff000000u;
        }
        value = (int32_t)tmp;
        break;

    case 4:
        value = decoder_read_int32(self_p);
        break;

    default:
        value = 2147483647;
        break;
    }

    return (value);
}\
'''

DECODER_READ_FLOAT = '''
static float decoder_read_float(struct decoder_t *self_p)
{
    float value;
    uint32_t i32;

    i32 = decoder_read_uint32(self_p);

    (void)memcpy(&value, &i32, sizeof(value));

    return (value);
}\
'''

DECODER_READ_DOUBLE = '''
static double decoder_read_double(struct decoder_t *self_p)
{
    double value;
    uint64_t i64;

    i64 = decoder_read_uint64(self_p);

    (void)memcpy(&value, &i64, sizeof(value));

    return (value);
}\
'''

DECODER_READ_BOOL = '''
static bool decoder_read_bool(struct decoder_t *self_p)
{
    return (decoder_read_uint8(self_p) != 0u);
}\
'''

DECODER_READ_LENGTH_DETERMINANT = '''
static uint32_t decoder_read_length_determinant(struct decoder_t *self_p)
{
    uint32_t length;

    length = decoder_read_uint8(self_p);

    if ((length & 0x80u) != 0u) {
        switch (length & 0x7fu) {

        case 1:
            length = decoder_read_uint8(self_p);
            break;

        case 2:
            length = decoder_read_uint16(self_p);
            break;

        case 3:
            length = (((uint32_t)decoder_read_uint8(self_p) << 16)
                      | decoder_read_uint16(self_p));
            break;

        case 4:
            length = decoder_read_uint32(self_p);
            break;

        default:
            length = 0xffffffffu;
            break;
        }
    }

    return (length);
}\
'''

DECODER_READ_TAG = '''
static uint32_t decoder_read_tag(struct decoder_t *self_p)
{
    uint32_t tag;

    tag = decoder_read_uint8(self_p);

    if ((tag & 0x3fu) == 0x3fu) {
        do {
            tag <<= 8;
            tag |= (uint32_t)decoder_read_uint8(self_p);
        } while ((tag & 0x80u) == 0x80u);
    }

    return (tag);
}\
'''

functions = [
    ('decoder_read_tag(', DECODER_READ_TAG),
    ('decoder_read_length_determinant(', DECODER_READ_LENGTH_DETERMINANT),
    ('decoder_read_bool(', DECODER_READ_BOOL),
    ('decoder_read_double(', DECODER_READ_DOUBLE),
    ('decoder_read_float(', DECODER_READ_FLOAT),
    ('decoder_read_int(', DECODER_READ_INT),
    ('decoder_read_uint(', DECODER_READ_UINT),
    ('decoder_read_long_uint(', DECODER_READ_LONG_UINT),
    ('decoder_read_int64(', DECODER_READ_INT64),
    ('decoder_read_int32(', DECODER_READ_INT32),
    ('decoder_read_int16(', DECODER_READ_INT16),
    ('decoder_read_int8(', DECODER_READ_INT8),
    ('decoder_read_uint64(', DECODER_READ_UINT64),
    ('decoder_read_uint32(', DECODER_READ_UINT32),
    ('decoder_read_uint16(', DECODER_READ_UINT16),
    ('decoder_read_uint8(', DECODER_READ_UINT8),
    ('decoder_read_bytes(', DECODER_READ_BYTES),
    ('decoder_free(', DECODER_FREE),
    ('decoder_abort(', DECODER_ABORT),
    ('decoder_get_result(', DECODER_GET_RESULT),
    ('decoder_init(', DECODER_INIT),
    ('encoder_append_length_determinant(', ENCODER_APPEND_LENGTH_DETERMINANT),
    ('encoder_append_bool(', ENCODER_APPEND_BOOL),
    ('encoder_append_double(', ENCODER_APPEND_DOUBLE),
    ('encoder_append_float(', ENCODER_APPEND_FLOAT),
    ('encoder_append_int(', ENCODER_APPEND_INT),
    ('encoder_append_uint(', ENCODER_APPEND_UINT),
    ('encoder_append_long_uint(', ENCODER_APPEND_LONG_UINT),
    ('encoder_append_int64(', ENCODER_APPEND_INT64),
    ('encoder_append_int32(', ENCODER_APPEND_INT32),
    ('encoder_append_int16(', ENCODER_APPEND_INT16),
    ('encoder_append_int8(', ENCODER_APPEND_INT8),
    ('encoder_append_uint64(', ENCODER_APPEND_UINT64),
    ('encoder_append_uint32(', ENCODER_APPEND_UINT32),
    ('encoder_append_uint16(', ENCODER_APPEND_UINT16),
    ('encoder_append_uint8(', ENCODER_APPEND_UINT8),
    ('encoder_append_bytes(', ENCODER_APPEND_BYTES),
    ('encoder_alloc(', ENCODER_ALLOC),
    ('encoder_abort(', ENCODER_ABORT),
    ('encoder_get_result(', ENCODER_GET_RESULT),
    ('encoder_init(', ENCODER_INIT),
    ('minimum_uint_length(', MINIMUM_UINT_LENGTH),
    ('length_determinant_length(', LENGTH_DETERMINANT_LENGTH),
    ('enumerated_value_length(', ENUMERATED_VALUE_LENGTH)
]



import re

from ...errors import Error


TYPE_DECLARATION_FMT = '''\
/// Type {type_name} in module {module_name}.
{members}
'''

DEFINITION_FMT = '''
impl {module_name}{type_name} {{
    pub fn encode(&mut self, mut dst: &mut [u8]) -> Result<usize, Error> {{
        let mut encoder = Encoder::new(&mut dst);

        self.encode_inner(&mut encoder);

        encoder.get_result()
    }}

    pub fn decode(&mut self, src: &[u8]) -> Result<usize, Error> {{
        let mut decoder = Decoder::new(&src);

        self.decode_inner(&mut decoder);

        decoder.get_result()
    }}

    fn encode_inner(&mut self, encoder: &mut Encoder) {{
{encode_body}\
    }}

    fn decode_inner(&mut self, decoder: &mut Decoder) {{
{decode_body}\
    }}
}}
'''

ENCODER_AND_DECODER_STRUCTS = '''\
#[derive(Debug, PartialEq, Copy, Clone)]
pub enum Error {
    BadChoice,
    BadEnum,
    BadLength,
    OutOfData,
    OutOfMemory
}

struct Encoder<'a> {
    buf: &'a mut [u8],
    size: usize,
    pos: usize,
    error: Option<Error>
}

struct Decoder<'a> {
    buf: &'a[u8],
    size: usize,
    pos: usize,
    error: Option<Error>
}
'''

ENCODER_ABORT = '''
    fn abort(&mut self, error: Error) {
        if self.error.is_none() {
            self.error = Some(error);
        }
    }\
'''

DECODER_ABORT = '''
    fn abort(&mut self, error: Error) {
        if self.error.is_none() {
            self.error = Some(error);
        }
    }\
'''


class _MembersBacktracesContext(object):

    def __init__(self, backtraces, member_name):
        self.backtraces = backtraces
        self.member_name = member_name

    def __enter__(self):
        for backtrace in self.backtraces:
            backtrace.append(self.member_name)

    def __exit__(self, *args):
        for backtrace in self.backtraces:
            backtrace.pop()


class _UserType(object):

    def __init__(self,
                 type_name,
                 module_name,
                 type_code,
                 used_user_types):
        self.type_name = type_name
        self.module_name = module_name
        self.type_code = type_code
        self.used_user_types = used_user_types


class Generator(object):

    def __init__(self):
        self.namespace = 'a'
        self.asn1_members_backtrace = []
        self.c_members_backtrace = []
        self.module_name = None
        self.type_name = None
        self.helper_lines = []
        self.base_variables = set()
        self.used_suffixes_by_base_variables = {}
        self.encode_variable_lines = []
        self.decode_variable_lines = []
        self.used_user_types = []

    def reset_type(self):
        self.helper_lines = []
        self.base_variables = set()
        self.used_suffixes_by_base_variables = {}
        self.encode_variable_lines = []
        self.decode_variable_lines = []
        self.used_user_types = []

    @property
    def module_name_snake(self):
        return camel_to_snake_case(self.module_name)

    @property
    def type_name_snake(self):
        return camel_to_snake_case(self.type_name)

    def type_length(self, minimum, maximum):
        # Make sure it fits in 64 bits.
        if minimum < -9223372036854775808:
            raise self.error(
                '{} does not fit in int64_t.'.format(minimum))
        elif maximum > 18446744073709551615:
            raise self.error(
                '{} does not fit in uint64_t.'.format(maximum))
        elif minimum < 0 and maximum > 9223372036854775807:
            raise self.error(
                '{} does not fit in int64_t.'.format(maximum))

        # Calculate the number of bytes needed.
        if minimum < -4294967296:
            minimum_length = 64
        elif minimum < -65536:
            minimum_length = 32
        elif minimum < -256:
            minimum_length = 16
        elif minimum < 0:
            minimum_length = 8
        else:
            minimum_length = 0

        if maximum > 4294967295:
            maximum_length = 64
        elif maximum > 65535:
            maximum_length = 32
        elif maximum > 255:
            maximum_length = 16
        elif maximum > 0:
            maximum_length = 8
        else:
            maximum_length = 0

        if minimum_length == maximum_length == 0:
            length = 8
        else:
            length = max(minimum_length, maximum_length)

        return length

    def format_type_name(self, minimum, maximum):
        length = self.type_length(minimum, maximum)

        if minimum >= 0:
            type_name = 'u{}'.format(length)
        else:
            type_name = 'i{}'.format(length)

        return type_name

    @property
    def location(self):
        location = '{}{}'.format(self.module_name,
                                 self.type_name)

        for member in self.asn1_members_backtrace:
            location += make_camel_case(member)

        return location

    def location_inner(self, default='value', end=''):
        if self.c_members_backtrace:
            return '.'.join(self.c_members_backtrace) + end
        else:
            return default

    def location_error(self):
        location = '{}.{}'.format(self.module_name, self.type_name)

        if self.asn1_members_backtrace:
            location += '.{}'.format('.'.join(self.asn1_members_backtrace))

        return location

    def members_backtrace_push(self, member_name):
        backtraces = [
            self.asn1_members_backtrace,
            self.c_members_backtrace
        ]

        return _MembersBacktracesContext(backtraces, member_name)

    def asn1_members_backtrace_push(self, member_name):
        backtraces = [self.asn1_members_backtrace]

        return _MembersBacktracesContext(backtraces, member_name)

    def c_members_backtrace_push(self, member_name):
        backtraces = [self.c_members_backtrace]

        return _MembersBacktracesContext(backtraces, member_name)

    def get_member_checker(self, checker, name):
        for member in checker.members:
            if member.name == name:
                return member

        raise Error('No member checker found for {}.'.format(name))

    def add_unique_variable(self, name):
        if name in self.base_variables:
            try:
                suffix = self.used_suffixes_by_base_variables[name]
                suffix += 1
            except KeyError:
                suffix = 2

            self.used_suffixes_by_base_variables[name] = suffix
            unique_name = '{}_{}'.format(name, suffix)
        else:
            self.base_variables.add(name)
            unique_name = name

        return unique_name

    def error(self, message):
        return Error('{}: {}'.format(self.location_error(), message))

    def format_integer(self, checker):
        if not checker.has_lower_bound():
            raise self.error('INTEGER has no minimum value.')

        if not checker.has_upper_bound():
            raise self.error('INTEGER has no maximum value.')

        type_name = self.format_type_name(checker.minimum, checker.maximum)

        return [type_name]

    def format_boolean(self):
        return ['bool']

    def format_octet_string(self, checker):
        if not checker.has_upper_bound():
            raise self.error('OCTET STRING has no maximum length.')

        if checker.minimum == checker.maximum:
            lines = []
        elif checker.maximum < 256:
            lines = ['    let length: u8;']
        else:
            lines = ['    let length: u32;']

        return [
            '#[derive(Debug, Default, PartialEq, Copy, Clone)]',
            'pub struct {} {{'.format(self.location)
        ] + lines + [
            '    pub buf: [u8; {}]'.format(checker.maximum),
            '}'
        ]

    def format_sequence(self, type_, checker):
        helper_lines = []
        lines = []

        for member in type_.root_members:
            member_checker = self.get_member_checker(checker, member.name)

            if member.optional:
                lines += ['pub is_{}_present: bool,'.format(member.name)]

            with self.members_backtrace_push(member.name):
                member_lines = self.format_type(member, member_checker)
                member_location = self.location

            if not member_lines:
                continue

            if is_inline_member_lines(member_lines):
                member_lines[-1] = 'pub {}: {},'.format(member.name,
                                                        member_lines[-1])
            else:
                helper_lines += member_lines + ['']
                member_lines = ['pub {}: {},'.format(member.name,
                                                     member_location)]

            lines += member_lines

        if lines:
            lines[-1] = lines[-1].strip(',')

        return helper_lines + [
            '#[derive(Debug, Default, PartialEq, Copy, Clone)]',
            'pub struct {} {{'.format(self.location)
        ] + indent_lines(lines) + [
            '}'
        ]

    def format_sequence_of(self, type_, checker):
        if not checker.is_bound():
            raise self.error('SEQUENCE OF has no maximum length.')

        with self.asn1_members_backtrace_push('elem'):
            lines = self.format_type(type_.element_type,
                                     checker.element_type)

        if lines:
            lines[-1] += ' elements[{}];'.format(checker.maximum)

        if checker.minimum == checker.maximum:
            length_lines = []
        elif checker.maximum < 256:
            length_lines = ['let length: u8;']
        else:
            length_lines = ['let length: u32;']

        return ['struct {'] + indent_lines(length_lines + lines) + ['}']

    def format_enumerated(self, type_):
        lines = [
            '#[derive(Debug, PartialEq, Copy, Clone)]',
            'pub enum {} {{'.format(self.location)
        ] + [
            '    {},'.format(make_camel_case(value))
            for value in self.get_enumerated_values(type_)
        ] + [
            '}',
            '',
            'impl Default for {} {{'.format(self.location),
            '    fn default() -> Self {',
            '        {}::{}'.format(self.location,
                                    self.get_enumerated_values(type_)[0].upper()),
            '    }',
            '}'

        ]

        return lines

    def format_choice(self, type_, checker):
        helper_lines = []
        lines = []

        for member in self.get_choice_members(type_):
            member_checker = self.get_member_checker(checker,
                                                     member.name)

            with self.members_backtrace_push(member.name):
                member_lines = self.format_type(member, member_checker)
                member_location = self.location

            if not member_lines:
                continue

            if is_inline_member_lines(member_lines):
                member_lines[-1] = '{}({}),'.format(make_camel_case(member.name),
                                                    member_lines[-1])
            else:
                helper_lines += member_lines + ['']
                member_lines = ['pub {}: {},'.format(member.name,
                                                     member_location)]

            lines += member_lines

        if lines:
            lines[-1] = lines[-1].strip(',')

        return helper_lines + [
            '#[derive(Debug, PartialEq, Copy, Clone)]',
            'pub enum {} {{'.format(self.location)
        ] + indent_lines(lines) + [
            '}'
        ]

    def format_user_type(self, type_name, module_name):
        self.used_user_types.append((type_name, module_name))

        return ['{}{}'.format(module_name, type_name)]

    def format_sequence_inner_member(self,
                                     member,
                                     checker,
                                     default_condition_by_member_name):
        member_checker = self.get_member_checker(checker, member.name)

        with self.members_backtrace_push(member.name):
            encode_lines, decode_lines = self.format_type_inner(
                member,
                member_checker)

        location = self.location_inner('', '.')

        if member.optional:
            is_present = '{}is_{}_present'.format(location, member.name)
            encode_lines = [
                '',
                'if src.{} {{'.format(is_present)
            ] + indent_lines(encode_lines) + [
                '}',
                ''
            ]
            decode_lines = [
                '',
                'if dst.{} {{'.format(is_present)
            ] + indent_lines(decode_lines) + [
                '}',
                ''
            ]
        elif member.default is not None:
            name = '{}{}'.format(location, member.name)
            encode_lines = [
                '',
                'if src.{} != {} {{'.format(name, member.default)
            ] + indent_lines(encode_lines) + [
                '}',
                ''
            ]
            decode_lines = [
                '',
                'if {} {{'.format(default_condition_by_member_name[member.name])
            ] + indent_lines(decode_lines) + [
                '} else {',
                '    dst.{} = {};'.format(name, member.default),
                '}',
                ''
            ]

        return encode_lines, decode_lines

    def generate_type_declaration(self, compiled_type):
        type_ = compiled_type.type
        checker = compiled_type.constraints_checker.type

        lines = self.generate_type_declaration_process(type_, checker)

        if not lines:
            lines = ['dummy: u8;']

        if self.helper_lines:
            self.helper_lines.append('')

        return TYPE_DECLARATION_FMT.format(module_name=self.module_name,
                                           type_name=self.type_name,
                                           members='\n'.join(lines))

    def generate_definition(self, compiled_type):
        encode_lines, decode_lines = self.generate_definition_inner_process(
            compiled_type.type,
            compiled_type.constraints_checker.type)

        if self.encode_variable_lines:
            encode_lines = self.encode_variable_lines + [''] + encode_lines

        if self.decode_variable_lines:
            decode_lines = self.decode_variable_lines + [''] + decode_lines

        encode_lines = indent_lines(indent_lines(encode_lines)) + ['']
        decode_lines = indent_lines(indent_lines(decode_lines)) + ['']

        return DEFINITION_FMT.format(module_name=self.module_name,
                                     type_name=self.type_name,
                                     encode_body='\n'.join(encode_lines),
                                     decode_body='\n'.join(decode_lines))

    def generate(self, compiled):
        user_types = []

        for module_name, module in sorted(compiled.modules.items()):
            self.module_name = module_name

            for type_name, compiled_type in sorted(module.items()):
                self.type_name = type_name
                self.reset_type()

                type_declaration = self.generate_type_declaration(compiled_type)
                definition = self.generate_definition(compiled_type)
                user_type = _UserType(type_name,
                                      module_name,
                                      type_declaration + definition,
                                      self.used_user_types)
                user_types.append(user_type)

        user_types = sort_user_types_by_used_user_types(user_types)

        types_code = []

        for user_type in user_types:
            types_code.append(user_type.type_code)

        types_code = '\n'.join(types_code)
        helpers = '\n'.join(self.generate_helpers(types_code))

        return helpers, types_code

    def format_type(self, type_, checker):
        raise NotImplementedError('To be implemented by subclasses.')

    def format_type_inner(self, type_, checker):
        raise NotImplementedError('To be implemented by subclasses.')

    def get_enumerated_values(self, type_):
        raise NotImplementedError('To be implemented by subclasses.')

    def get_choice_members(self, type_):
        raise NotImplementedError('To be implemented by subclasses.')

    def generate_type_declaration_process(self, type_, checker):
        raise NotImplementedError('To be implemented by subclasses.')

    def generate_definition_inner_process(self, type_, checker):
        raise NotImplementedError('To be implemented by subclasses.')

    def generate_helpers(self, definitions):
        raise NotImplementedError('To be implemented by subclasses.')


def canonical(value):
    """Replace anything but 'a-z', 'A-Z' and '0-9' with '_'.

    """

    return re.sub(r'[^a-zA-Z0-9]', '_', value)


def camel_to_snake_case(value):
    value = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', value)
    value = re.sub(r'(_+)', '_', value)
    value = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', value).lower()
    value = canonical(value)

    return value


def make_camel_case(value):
    return value[0].upper() + value[1:]


def join_lines(lines, suffix):
    return[line + suffix for line in lines[:-1]] + lines[-1:]


def is_user_type(type_):
    return type_.module_name is not None


def strip_blank_lines(lines):
    try:
        while lines[0] == '':
            del lines[0]

        while lines[-1] == '':
            del lines[-1]
    except IndexError:
        pass

    stripped = []

    for line in lines:
        if line == '' and stripped[-1] == '':
            continue

        stripped.append(line)

    return stripped


def indent_lines(lines, width=4):
    indented_lines = []

    for line in lines:
        if line:
            indented_line = width * ' ' + line
        else:
            indented_line = line

        indented_lines.append(indented_line)

    return strip_blank_lines(indented_lines)


def dedent_lines(lines, width=4):
    return [line[width:] for line in lines]


def sort_user_types_by_used_user_types(user_types):
    reversed_sorted_user_types = []

    for user_type in user_types:
        user_type_name_tuple = (user_type.type_name, user_type.module_name)

        # Insert first in the reversed list if there are no types
        # using this type.
        insert_index = 0

        for i, reversed_sorted_user_type in enumerate(reversed_sorted_user_types, 1):
            if user_type_name_tuple in reversed_sorted_user_type.used_user_types:
                if i > insert_index:
                    insert_index = i

        reversed_sorted_user_types.insert(insert_index, user_type)

    return reversed(reversed_sorted_user_types)


def is_inline_member_lines(member_lines):
    return len(member_lines) == 1



EXPECTED = {'X683': {'extensibility-implied': False,
          'imports': {},
          'object-classes': {},
          'object-sets': {},
          'types': {'IntegerList1': {'members': [{'name': 'elem',
                                                  'type': 'INTEGER'},
                                                 {'name': 'next',
                                                  'optional': True,
                                                  'type': 'IntegerList1'}],
                                     'type': 'SEQUENCE'},
                    'OptionallySignedUTF8String': {'members': [{'name': 'unsigned-data',
                                                                'tag': {'number': 0},
                                                                'type': 'UTF8String'},
                                                               {'name': 'signed-data',
                                                                'tag': {'number': 1},
                                                                'type': 'SEQUENCE',
                                                                'members': [
                                                                    {'name': 'authenticated-data',
                                                                     'type': 'UTF8String'},
                                                                    {'name': 'authenticator',
                                                                     'type': 'BIT STRING'}]}],
                                                   'type': 'CHOICE'},
                    'SignedUTF8String': {'members': [{'name': 'authenticated-data',
                                                      'type': 'UTF8String'},
                                                     {'name': 'authenticator',
                                                      'type': 'BIT STRING'}],
                                         'type': 'SEQUENCE'}},
          'values': {}}}



import time

from ...version import __version__
from . import uper


SOURCE_FMT = '''\
//! This file was generated by asn1tools version {version} {date}.

{helpers}
{types_code}\
'''


def generate(compiled, codec):
    """Generate Rust source code from given compiled specification.

    """

    date = time.ctime()

    if codec == 'uper':
        helpers, types_code = uper.generate(compiled)
    else:
        raise Exception()

    source = SOURCE_FMT.format(version=__version__,
                               date=date,
                               helpers=helpers,
                               types_code=types_code)

    return source



import unittest
from datetime import datetime
from .utils import Asn1ToolsBaseTest
import asn1tools


class Asn1ToolsCheckConstraintsTest(Asn1ToolsBaseTest):

    maxDiff = None

    def assert_encode_decode_ok(self, foo, datas):
        for type_name, decoded in datas:
            encoded = foo.encode(type_name, decoded, check_constraints=True)
            foo.decode(type_name, encoded, check_constraints=True)

    def assert_encode_decode_bad(self, foo, datas, decode_check=True):
        for type_name, decoded, message in datas:
            # Encode check.
            with self.assertRaises(asn1tools.ConstraintsError) as cm:
                foo.encode(type_name, decoded, check_constraints=True)

            self.assertEqual(str(cm.exception), message)

            # Decode check.
            if decode_check:
                encoded = foo.encode(type_name, decoded, check_constraints=False)

                with self.assertRaises(asn1tools.ConstraintsError) as cm:
                    foo.decode(type_name, encoded, check_constraints=True)

                self.assertEqual(str(cm.exception), message)

    def test_all_codecs(self):
        codecs = [
            'ber',
            'der',
            'gser',
            'jer',
            'per',
            'uper',
            'xer'
        ]

        for codec in codecs:
            foo = asn1tools.compile_string(
                "Foo DEFINITIONS AUTOMATIC TAGS ::= "
                "BEGIN "
                "A ::= INTEGER "
                "END",
                codec)

            encoded = foo.encode('A', 0, check_constraints=True)

            if codec != 'gser':
                foo.decode('A', encoded, check_constraints=True)

    def test_boolean(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BOOLEAN "
            "END")

        # Ok.
        datas = [
            ('A', True),
            ('A', False)
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_integer(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= INTEGER "
            "B ::= INTEGER (5..99) "
            "C ::= INTEGER (-10..10) "
            "D ::= INTEGER (5..99, ...) "
            "E ::= INTEGER (1000..1000) "
            "F ::= SEQUENCE { "
            "  a INTEGER (4..4), "
            "  b INTEGER (40..40), "
            "  c INTEGER (400..400) "
            "} "
            "G ::= B (6..7) "
            "H ::= INTEGER (MIN..10) "
            "I ::= INTEGER (10..MAX) "
            "J ::= INTEGER (MIN..MAX) "
            "K ::= INTEGER (1..2 | 4..5) "
            "END")

        # Ok.
        datas = [
            ('A',  32768),
            ('A',      0),
            ('A', -32769),
            ('B',      5),
            ('B',      6),
            ('B',     99),
            ('C',    -10),
            ('C',     10),
            ('D',     99),
            ('E',   1000),
            ('F',   {'a': 4, 'b': 40, 'c': 400}),
            ('H',  -1000),
            ('H',     10),
            ('I',     10),
            ('I',   1000),
            ('J',  -1000),
            ('J',   1000),
            ('K',   1),
            ('K',   2)
        ]

        self.assert_encode_decode_ok(foo, datas)

        # ToDo: These should be good.
        datas = [
            ('K',
             4,
             'Expected an integer between 1 and 2, but got 4.'),
            ('K',
             5,
             'Expected an integer between 1 and 2, but got 5.')
        ]

        self.assert_encode_decode_bad(foo, datas)

        # Not ok.
        datas = [
            ('B',
             4,
             'Expected an integer between 5 and 99, but got 4.'),
            ('B',
             100,
             'Expected an integer between 5 and 99, but got 100.'),
            ('C',
             -11,
             'Expected an integer between -10 and 10, but got -11.'),
            ('C',
             11,
             'Expected an integer between -10 and 10, but got 11.'),
            ('E',
             0,
             'Expected an integer between 1000 and 1000, but got 0.'),
            ('F',
             {'a': 4, 'b': 41, 'c': 400},
             'b: Expected an integer between 40 and 40, but got 41.'),
            ('H',
             11,
             'Expected an integer between MIN and 10, but got 11.'),
            ('I',
             9,
             'Expected an integer between 10 and MAX, but got 9.'),
            # ToDo: 4..5 are allowed as well.
            ('K',
             3,
             'Expected an integer between 1 and 2, but got 3.')
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_real(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= REAL "
            "END")

        # Ok.
        datas = [
            ('A', 1.0)
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_null(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= NULL "
            "END")

        # Ok.
        datas = [
            ('A', None)
        ]

        self.assert_encode_decode_ok(foo, datas)

    def test_bit_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BIT STRING "
            "B ::= BIT STRING (SIZE (10)) "
            "END")

        # Ok.
        datas = [
            ('A',  (b'', 0)),
            ('B',  (b'\x01\x23', 10))
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('B',
             (b'\x01\x23', 9),
             'Expected between 10 and 10 bits, but got 9.')
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_octet_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= OCTET STRING "
            "B ::= OCTET STRING (SIZE (10)) "
            "C ::= OCTET STRING (SIZE (10, ...)) "
            "D ::= OCTET STRING (SIZE (10..MAX)) "
            "END")

        # Ok.
        datas = [
            ('A',  b''),
            ('B',  10 * b'\x23'),
            ('C',  11 * b'\x23'),
            ('D',  44 * b'\x23')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('B',
             11 * b'\x01',
             'Expected between 10 and 10 bytes, but got 11.'),
            ('D',
             9 * b'\x01',
             'Expected between 10 and MAX bytes, but got 9.')
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_enumerated(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= ENUMERATED { a, b } "
            "END")

        # Ok.
        datas = [
            ('A',  'a')
        ]

        self.assert_encode_decode_ok(foo, datas)

    def test_sequence(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a A OPTIONAL "
            "} "
            "END")

        # Ok.
        datas = [
            ('A',  {'a': {}})
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_sequence_of(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE (SIZE (2)) OF INTEGER (3..5)"
            "B ::= SEQUENCE (SIZE (2, ...)) OF INTEGER (3..5)"
            "C ::= SEQUENCE (SIZE (MIN..2)) OF INTEGER (3..5)"
            "END")

        # Ok.
        datas = [
            ('A',  [3, 4]),
            ('B',  [3, 4, 5]),
            ('C',  [3, 4])
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('A',
             [3, 4, 5],
             'Expected a list of between 2 and 2 elements, but got 3.'),
            ('A',
             [3, 6],
             'Expected an integer between 3 and 5, but got 6.'),
            ('C',
             [3, 4, 5],
             'Expected a list of between MIN and 2 elements, but got 3.')
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_numeric_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= NumericString "
            "END")

        # Ok.
        datas = [
            ('A',  '0123456789 ')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('A',
             'a',
             "Expected a character in ' 0123456789', but got 'a' (0x61)."),
            ('A',
             '-',
             "Expected a character in ' 0123456789', but got '-' (0x2d).")
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_printable_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= PrintableString "
            "END")

        # Ok.
        datas = [
            ('A',  '09azAZ')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('A',
             '{',
             "Expected a character in '"
             "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
             " '()+,-./:=?', but got '{' (0x7b).")
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_ia5_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= IA5String "
            "END")

        # Ok.
        datas = [
            ('A',  '09azAZ')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('A',
             b'\x81'.decode('latin-1'),
             "Expected a character in '................................ !\"#$%"
             "&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcde"
             "fghijklmnopqrstuvwxyz{|}~.', but got '.' (0x81).")
        ]

        self.assert_encode_decode_bad(foo, datas, decode_check=False)

    def test_visible_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= VisibleString "
            "B ::= VisibleString (SIZE (2..5)) "
            "C ::= VisibleString (FROM (\"a\"..\"j\" | \"u\"..\"w\")) "
            "D ::= VisibleString (SIZE (2..5, ...)) "
            "END")

        # Ok.
        datas = [
            ('A',  '123'),
            ('B',  '12'),
            ('B',  '12345'),
            ('C',  'abijuvw'),
            ('D',  '123456')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('B',
             '1',
             'Expected between 2 and 5 characters, but got 1.'),
            ('B',
             '123456',
             'Expected between 2 and 5 characters, but got 6.'),
            ('C',
             'k',
             "Expected a character in 'abcdefghijuvw', but got 'k' (0x6b).")
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_utf8_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTF8String "
            "B ::= UTF8String (SIZE (2..5)) "
            "END")

        # Ok.
        datas = [
            ('A',  'Hello'),
            ('B',  '12'),
            ('B',  '12345')
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('B',
             '1',
             'Expected between 2 and 5 characters, but got 1.'),
            ('B',
             '123456',
             'Expected between 2 and 5 characters, but got 6.')
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_utc_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTCTime "
            "END")

        # Ok.
        datas = [
            ('A',  datetime(1982, 1, 2, 12, 0))
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
        ]

        self.assert_encode_decode_bad(foo, datas)

    def test_choice(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= CHOICE { a INTEGER (1..2) } "
            "END")

        # Ok.
        datas = [
            ('A',  ('a', 1))
        ]

        self.assert_encode_decode_ok(foo, datas)

        # Not ok.
        datas = [
            ('A',
             ('a', 3),
             'a: Expected an integer between 1 and 2, but got 3.')
        ]

        self.assert_encode_decode_bad(foo, datas)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'RFC1155-SMI': {'extensibility-implied': False,
                 'imports': {},
                 'object-classes': {},
                 'object-sets': {},
                 'types': {'ApplicationSyntax': {'members': [{'name': 'address',
                                                              'type': 'NetworkAddress'},
                                                             {'name': 'counter',
                                                              'type': 'Counter'},
                                                             {'name': 'gauge',
                                                              'type': 'Gauge'},
                                                             {'name': 'ticks',
                                                              'type': 'TimeTicks'},
                                                             {'name': 'arbitrary',
                                                              'type': 'Opaque'}],
                                                 'type': 'CHOICE'},
                           'Counter': {'restricted-to': [(0, 4294967295)],
                                       'tag': {'class': 'APPLICATION',
                                               'kind': 'IMPLICIT',
                                               'number': 1},
                                       'type': 'INTEGER'},
                           'Gauge': {'restricted-to': [(0, 4294967295)],
                                     'tag': {'class': 'APPLICATION',
                                             'kind': 'IMPLICIT',
                                             'number': 2},
                                     'type': 'INTEGER'},
                           'IpAddress': {'size': [4],
                                         'tag': {'class': 'APPLICATION',
                                                 'kind': 'IMPLICIT',
                                                 'number': 0},
                                         'type': 'OCTET STRING'},
                           'NetworkAddress': {'members': [{'name': 'internet',
                                                           'type': 'IpAddress'}],
                                              'type': 'CHOICE'},
                           'ObjectName': {'type': 'OBJECT IDENTIFIER'},
                           'ObjectSyntax': {'members': [{'name': 'simple',
                                                         'type': 'SimpleSyntax'},
                                                        {'name': 'application-wide',
                                                         'type': 'ApplicationSyntax'}],
                                            'type': 'CHOICE'},
                           'Opaque': {'tag': {'class': 'APPLICATION',
                                              'kind': 'IMPLICIT',
                                              'number': 4},
                                      'type': 'OCTET STRING'},
                           'SimpleSyntax': {'members': [{'name': 'number',
                                                         'type': 'INTEGER'},
                                                        {'name': 'string',
                                                         'type': 'OCTET '
                                                                 'STRING'},
                                                        {'name': 'object',
                                                         'type': 'OBJECT '
                                                                 'IDENTIFIER'},
                                                        {'name': 'empty',
                                                         'type': 'NULL'}],
                                            'type': 'CHOICE'},
                           'TimeTicks': {'restricted-to': [(0, 4294967295)],
                                         'tag': {'class': 'APPLICATION',
                                                 'kind': 'IMPLICIT',
                                                 'number': 3},
                                         'type': 'INTEGER'}},
                 'values': {'directory': {'type': 'OBJECT IDENTIFIER',
                                          'value': ['internet', 1]},
                            'enterprises': {'type': 'OBJECT IDENTIFIER',
                                            'value': ['private', 1]},
                            'experimental': {'type': 'OBJECT IDENTIFIER',
                                             'value': ['internet', 3]},
                            'internet': {'type': 'OBJECT IDENTIFIER',
                                         'value': ['iso',
                                                   ('org', 3),
                                                   ('dod', 6),
                                                   1]},
                            'mgmt': {'type': 'OBJECT IDENTIFIER',
                                     'value': ['internet', 2]},
                            'private': {'type': 'OBJECT IDENTIFIER',
                                        'value': ['internet', 4]}}}}


"""General String Encoding Rules (GSER).

"""

import time
import binascii
import math
from copy import copy
import datetime

from . import EncodeError
from . import DecodeError
from . import compiler
from . import format_or
from . import utc_time_from_datetime
from . import generalized_time_from_datetime
from .compiler import enum_values_as_dict


class Type(object):

    def __init__(self, name, type_name):
        self.name = name
        self.type_name = type_name
        self.optional = False
        self.default = None

    def set_size_range(self, minimum, maximum, has_extension_marker):
        pass

    def set_default(self, value):
        self.default = value

    def has_default(self):
        return self.default is not None


class MembersType(Type):

    def __init__(self, name, members, type_name):
        super(MembersType, self).__init__(name, type_name)
        self.members = members

    def encode(self, data, separator, indent):
        encoded_members = []
        member_separator = separator + ' ' * indent

        for member in self.members:
            name = member.name

            if name in data:
                try:
                    encoded_member = member.encode(data[name],
                                                   member_separator,
                                                   indent)
                except EncodeError as e:
                    e.location.append(member.name)
                    raise

                encoded_member = u'{}{} {}'.format(member_separator,
                                                   member.name,
                                                   encoded_member)
                encoded_members.append(encoded_member)
            elif member.optional:
                pass
            elif not member.has_default():
                raise EncodeError(
                    "{} member '{}' not found in {}.".format(
                        self.__class__.__name__,
                        name,
                        data))

        encoded_members = ','.join(encoded_members)

        return separator.join(['{' + encoded_members, '}'])

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.members]))


class ArrayType(Type):

    def __init__(self, name, type_name, element_type):
        super(ArrayType, self).__init__(name, type_name)
        self.element_type = element_type

    def encode(self, data, separator, indent):
        encoded_elements = []
        element_separator = separator + ' ' * indent

        for entry in data:
            encoded_element = self.element_type.encode(entry,
                                                       element_separator,
                                                       indent)
            encoded_element = u'{}{}'.format(element_separator,
                                             encoded_element)
            encoded_elements.append(encoded_element)

        encoded_elements = ','.join(encoded_elements)

        return separator.join(['{' + encoded_elements, '}'])

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name, 'BOOLEAN')

    def encode(self, data, _separator, _indent):
        return 'TRUE' if data else 'FALSE'

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name, 'INTEGER')

    def encode(self, data, _separator, _indent):
        return str(data)

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL')

    def encode(self, data, _separator, _indent):
        if data == float('inf'):
            data = 'PLUS-INFINITY'
        elif data == float('-inf'):
            data = 'MINUS-INFINITY'
        elif math.isnan(data):
            raise EncodeError('Cannot encode floating point number NaN.')
        elif data == 0.0:
            data = '0'
        else:
            data = '{}E0'.format(data)

        return data

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL')

    def encode(self, _data, _separator, _indent):
        return 'NULL'

    def __repr__(self):
        return 'Null({})'.format(self.name)


class BitString(Type):

    def __init__(self, name):
        super(BitString, self).__init__(name, 'BIT STRING')

    def encode(self, data, _separator, _indent):
        encoded = int(binascii.hexlify(data[0]), 16)
        encoded |= (0x80 << (8 * len(data[0])))

        return "'{}'B".format(bin(encoded)[10:10 + data[1]]).upper()

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name, 'OCTET STRING')

    def encode(self, data, _separator, _indent):
        return "'{}'H".format(binascii.hexlify(data).decode('ascii')).upper()

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name, 'OBJECT IDENTIFIER')

    def encode(self, data, _separator, _indent):
        return data

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values, numeric):
        super(Enumerated, self).__init__(name, 'ENUMERATED')

        if numeric:
            self.data_to_value = enum_values_as_dict(values)
        else:
            self.data_to_value = {
                v: v for v in enum_values_as_dict(values).values()
            }

    def encode(self, data, _separator, _indent):
        return self.data_to_value[data]

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class Sequence(MembersType):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, members, 'SEQUENCE')


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         'SEQUENCE OF',
                                         element_type)


class Set(MembersType):

    def __init__(self, name, members):
        super(Set, self).__init__(name, members, 'SET')


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    'SET OF',
                                    element_type)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE')
        self.members = members
        self.name_to_member = {member.name: member for member in self.members}

    def format_names(self):
        return format_or(sorted([member.name for member in self.members]))

    def encode(self, data, separator, indent):
        try:
            member = self.name_to_member[data[0]]
        except KeyError:
            raise EncodeError(
                "Expected choice {}, but got '{}'.".format(
                    self.format_names(),
                    data[0]))

        try:
            encoded = member.encode(data[1], separator, indent)
        except EncodeError as e:
            e.location.append(member.name)
            raise

        return u'{} : {}'.format(data[0], encoded)

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class UTF8String(Type):

    def __init__(self, name):
        super(UTF8String, self).__init__(name, 'UTF8String')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class NumericString(Type):

    def __init__(self, name):
        super(NumericString, self).__init__(name, 'NumericString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'NumericString({})'.format(self.name)


class PrintableString(Type):

    def __init__(self, name):
        super(PrintableString, self).__init__(name, 'PrintableString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'PrintableString({})'.format(self.name)


class IA5String(Type):

    def __init__(self, name):
        super(IA5String, self).__init__(name, 'IA5String')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'IA5String({})'.format(self.name)


class VisibleString(Type):

    def __init__(self, name):
        super(VisibleString, self).__init__(name, 'VisibleString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'VisibleString({})'.format(self.name)


class GeneralString(Type):

    def __init__(self, name):
        super(GeneralString, self).__init__(name, 'GeneralString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class BMPString(Type):

    def __init__(self, name):
        super(BMPString, self).__init__(name, 'BMPString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class GraphicString(Type):

    def __init__(self, name):
        super(GraphicString, self).__init__(name, 'GraphicString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'GraphicString({})'.format(self.name)


class UniversalString(Type):

    def __init__(self, name):
        super(UniversalString, self).__init__(name, 'UniversalString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class TeletexString(Type):

    def __init__(self, name):
        super(TeletexString, self).__init__(name, 'TeletexString')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(data)

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class ObjectDescriptor(GraphicString):

    def __repr__(self):
        return 'ObjectDescriptor({})'.format(self.name)


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name, 'UTCTime')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(utc_time_from_datetime(data))

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name, 'GeneralizedTime')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(generalized_time_from_datetime(data))

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class Date(Type):

    def __init__(self, name):
        super(Date, self).__init__(name, 'DATE')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(str(data))


class TimeOfDay(Type):

    def __init__(self, name):
        super(TimeOfDay, self).__init__(name, 'TIME-OF-DAY')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(str(data))


class DateTime(Type):

    def __init__(self, name):
        super(DateTime, self).__init__(name, 'DATE-TIME')

    def encode(self, data, _separator, _indent):
        return u'"{}"'.format(str(data).replace(' ', 'T'))


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY')

    def encode(self, data, _separator, _indent):
        data = binascii.hexlify(data).decode('ascii').upper()

        return "'{}'H".format(data)

    def __repr__(self):
        return 'Any({})'.format(self.name)


class Recursive(Type, compiler.Recursive):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE')
        self.type_name = type_name
        self.module_name = module_name
        self.inner = None

    def set_inner_type(self, inner):
        self.inner = copy(inner)

    def encode(self, data, separator, indent):
        return self.inner.encode(data, separator, indent)

    def __repr__(self):
        return 'Recursive({})'.format(self.type_name)


class CompiledType(compiler.CompiledType):

    def __init__(self, type_name, compiled_type):
        super(CompiledType, self).__init__(compiled_type)
        self._value_name = type_name.lower()
        self._value_type = type_name

    def encode(self, data, indent=None):
        if indent is None:
            encoded = self._type.encode(data, ' ', 0)
        else:
            encoded = self._type.encode(data, '\n', indent)

        encoded = u'{} {} ::= {}'.format(self._value_name,
                                         self._value_type,
                                         encoded.lstrip(' '))

        return encoded.encode('utf-8')

    def decode(self, data):
        raise NotImplementedError('GSER decoding is not implemented.')


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        compiled_type = self.compile_type(type_name,
                                          type_descriptor,
                                          module_name)

        return CompiledType(type_name, compiled_type)

    def compile_type(self, name, type_descriptor, module_name):
        module_name = self.get_module_name(type_descriptor, module_name)
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            members, _ = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Sequence(name, members)
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            members, _ = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Set(name, members)
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            members, _ = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Choice(name, members)
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name,
                                  self.get_enum_values(type_descriptor,
                                                       module_name),
                                  self._numeric_enums)
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'DATE':
            compiled = Date(name)
        elif type_name == 'TIME-OF-DAY':
            compiled = TimeOfDay(name)
        elif type_name == 'DATE-TIME':
            compiled = DateTime(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        elif type_name == 'EXTERNAL':
            members, _ = self.compile_members(
                self.external_type_descriptor()['members'],
                module_name)
            compiled = Sequence(name, members)
        elif type_name == 'ObjectDescriptor':
            compiled = ObjectDescriptor(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        return compiled


def compile_dict(specification, numeric_enums=False):
    return Compiler(specification, numeric_enums).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')



EXPECTED = {'AllTypes': {'extensibility-implied': False,
              'imports': {},
              'object-classes': {},
              'object-sets': {},
              'types': {'Any': {'type': 'ANY'},
                        'Bitstring': {'type': 'BIT STRING'},
                        'Bitstring2': {'size': [9], 'type': 'BIT STRING'},
                        'Bitstring3': {'size': [(5, 7)], 'type': 'BIT STRING'},
                        'Bitstring4': {'named-bits': [('a', '0'),
                                                      ('b', '1'),
                                                      ('c', '2')],
                                       'type': 'BIT STRING'},
                        'Bitstring5': {'named-bits': [('a', '0'),
                                                      ('b', '1'),
                                                      ('c', '2')],
                                       'size': [(2, 3)],
                                       'type': 'BIT STRING'},
                        'Bitstring6': {'size': [(1, 2), None],
                                       'type': 'BIT STRING'},
                        'Bmpstring': {'type': 'BMPString'},
                        'Boolean': {'type': 'BOOLEAN'},
                        'Boolean2': {'restricted-to': ['FALSE'],
                                     'type': 'BOOLEAN'},
                        'Boolean3': {'restricted-to': ['TRUE', None],
                                     'type': 'BOOLEAN'},
                        'Choice': {'members': [{'name': 'a',
                                                'type': 'INTEGER'}],
                                   'type': 'CHOICE'},
                        'Enumerated': {'type': 'ENUMERATED',
                                       'values': [('one', 1)]},
                        'Enumerated2': {'type': 'ENUMERATED',
                                        'values': [('one', 1),
                                                   ('two', 2),
                                                   None,
                                                   ('three', 3)]},
                        'GeneralizedTime1': {'type': 'GeneralizedTime'},
                        'Generalstring': {'type': 'GeneralString'},
                        'Graphicstring': {'type': 'GraphicString'},
                        'Ia5string': {'type': 'IA5String'},
                        'Ia5string2': {'from': [('a', 'f'), ('0', '9')],
                                       'type': 'IA5String'},
                        'Ia5string3': {'restricted-to': ['foo',
                                                         'bar',
                                                         'fie',
                                                         '...'],
                                       'type': 'IA5String'},
                        'Ia5string4': {'from': [('0', '2')],
                                       'size': [1],
                                       'type': 'IA5String'},
                        'Integer': {'type': 'INTEGER'},
                        'Integer2': {'restricted-to': [(1, 99)],
                                     'type': 'INTEGER'},
                        'Integer3': {'restricted-to': [(1, 2), (9, 10)],
                                     'type': 'INTEGER'},
                        'Integer4': {'size': [4], 'type': 'INTEGER'},
                        'Integer5': {'named-numbers': {'a': -1, 'b': 0, 'c': 1},
                                     'type': 'INTEGER'},
                        'Integer6': {'restricted-to': [(4, 7), None],
                                     'type': 'INTEGER'},
                        'List': {'element': {'type': 'INTEGER'},
                                 'type': 'SEQUENCE OF'},
                        'Null': {'type': 'NULL'},
                        'Numericstring': {'type': 'NumericString'},
                        'Objectidentifier': {'type': 'OBJECT IDENTIFIER'},
                        'Octetstring': {'type': 'OCTET STRING'},
                        'Octetstring2': {'size': [2], 'type': 'OCTET STRING'},
                        'Octetstring3': {'size': [3], 'type': 'OCTET STRING'},
                        'Octetstring4': {'size': [(3, 7)],
                                         'type': 'OCTET STRING'},
                        'Octetstring5': {'size': [4, 9],
                                         'type': 'OCTET STRING'},
                        'Printablestring': {'type': 'PrintableString'},
                        'Real': {'type': 'REAL'},
                        'Real10': {'type': 'REAL',
                                   'with-components': [('mantissa',
                                                        (-16777215, 16777215)),
                                                       ('base', 2),
                                                       ('exponent',
                                                        (-149, 104))]},
                        'Real2': {'restricted-to': [('1.0', '2.0')],
                                  'type': 'REAL'},
                        'Real3': {'restricted-to': [(1, 2)], 'type': 'REAL'},
                        'Real4': {'restricted-to': [('1.', 2)], 'type': 'REAL'},
                        'Real5': {'restricted-to': [('1.', 2)], 'type': 'REAL'},
                        'Real6': {'restricted-to': [('-20.0', '-10.0')],
                                  'type': 'REAL'},
                        'Real7': {'tag': {'class': 'UNIVERSAL', 'number': 0},
                                  'type': 'REAL'},
                        'Real8': {'tag': {'class': 'APPLICATION', 'number': 0},
                                  'type': 'REAL'},
                        'Real9': {'tag': {'class': 'PRIVATE', 'number': 0},
                                  'type': 'REAL'},
                        'Sequence': {'members': [], 'type': 'SEQUENCE'},
                        'Sequence10': {'members': [{'name': 'c',
                                                    'type': 'Sequence9',
                                                    'with-components': [None,
                                                                        ('b',
                                                                         'ABSENT')]},
                                                   {'name': 'd',
                                                    'type': 'Sequence9'},
                                                   None],
                                       'type': 'SEQUENCE'},
                        'Sequence11': {'type': 'Sequence10',
                                       'with-components': [['c',
                                                            [{'with-components': [('a',
                                                                                   32)]}],
                                                            'OPTIONAL'],
                                                           ['d']]},
                        'Sequence12': {'members': [{'element': {'type': 'Sequence12'},
                                                    'name': 'a',
                                                    'optional': True,
                                                    'type': 'SEQUENCE OF'}],
                                       'type': 'SEQUENCE'},
                        'Sequence13': {'members': [{'name': 'a',
                                                    'optional': True,
                                                    'tag': {'kind': 'IMPLICIT',
                                                            'number': 0},
                                                    'type': 'List'},
                                                   {'name': 'b',
                                                    'optional': True,
                                                    'tag': {'kind': 'IMPLICIT',
                                                            'number': 1},
                                                    'type': 'List'}],
                                       'type': 'SEQUENCE'},
                        'Sequence14': {'members': [{'name': 'a',
                                                    'tag': {'number': 0},
                                                    'type': 'INTEGER'},
                                                   {'name': 'b',
                                                    'tag': {'number': 1},
                                                    'type': 'Choice'},
                                                   {'name': 'c',
                                                    'tag': {'number': 2},
                                                    'type': 'Sequence9'},
                                                   {'members': [{'name': 'a',
                                                                 'type': 'BOOLEAN'}],
                                                    'name': 'd',
                                                    'tag': {'number': 3},
                                                    'type': 'CHOICE'}],
                                       'type': 'SEQUENCE'},
                        'Sequence15': {'members': [{'default': ['a', 'b'],
                                                    'name': 'a',
                                                    'type': 'Bitstring4'},
                                                   {'default': '0b110',
                                                    'name': 'b',
                                                    'type': 'Bitstring4'},
                                                   {'default': '0xc',
                                                    'name': 'c',
                                                    'type': 'Bitstring4'}],
                                       'type': 'SEQUENCE'},
                        'Sequence16': {'type': 'Sequence9',
                                       'with-components': [('b', 'PRESENT')]},
                        'Sequence17': {'type': 'Sequence9',
                                       'with-components': [('b', 'ABSENT')]},
                        'Sequence18': {'restricted-to': [{'type': 'Sequence16'},
                                                         {'type': 'Sequence17'}],
                                       'type': 'Sequence9'},
                        'Sequence2': {'members': [{'default': 0,
                                                   'name': 'a',
                                                   'type': 'INTEGER'}],
                                      'type': 'SEQUENCE'},
                        'Sequence3': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None],
                                      'type': 'SEQUENCE'},
                        'Sequence4': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None,
                                                  [{'name': 'b',
                                                    'type': 'BOOLEAN'}]],
                                      'type': 'SEQUENCE'},
                        'Sequence5': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None,
                                                  [{'name': 'b',
                                                    'type': 'BOOLEAN'}],
                                                  None],
                                      'type': 'SEQUENCE'},
                        'Sequence6': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None,
                                                  [{'name': 'b',
                                                    'type': 'BOOLEAN'}],
                                                  None,
                                                  {'name': 'c',
                                                   'type': 'BOOLEAN'}],
                                      'type': 'SEQUENCE'},
                        'Sequence7': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None,
                                                  [{'name': 'b',
                                                    'type': 'BOOLEAN'}],
                                                  [{'name': 'c',
                                                    'type': 'BOOLEAN'}],
                                                  None,
                                                  {'name': 'd',
                                                   'type': 'BOOLEAN'}],
                                      'type': 'SEQUENCE'},
                        'Sequence8': {'members': [{'name': 'a',
                                                   'type': 'BOOLEAN'},
                                                  None,
                                                  None],
                                      'type': 'SEQUENCE'},
                        'Sequence9': {'members': [{'name': 'a',
                                                   'type': 'INTEGER'},
                                                  {'name': 'b',
                                                   'optional': True,
                                                   'type': 'BOOLEAN'}],
                                      'type': 'SEQUENCE'},
                        'SequenceOf': {'element': {'type': 'INTEGER'},
                                       'type': 'SEQUENCE OF'},
                        'Set': {'members': [], 'type': 'SET'},
                        'Set2': {'members': [{'default': 1,
                                              'name': 'a',
                                              'type': 'INTEGER'}],
                                 'type': 'SET'},
                        'SetOf': {'element': {'type': 'INTEGER'},
                                  'type': 'SET OF'},
                        'Teletexstring': {'type': 'TeletexString'},
                        'Universalstring': {'type': 'UniversalString'},
                        'Utctime': {'type': 'UTCTime'},
                        'Utf8string': {'type': 'UTF8String'},
                        'Visiblestring': {'type': 'VisibleString'}},
              'values': {'bitstring': {'type': 'BIT STRING',
                                       'value': '0b100010001001'},
                         'bitstring2': {'type': 'BIT STRING',
                                        'value': '0x0123456789abcdef'},
                         'boolean': {'type': 'BOOLEAN', 'value': True},
                         'boolean2': {'type': 'BOOLEAN', 'value': False},
                         'choice': {'type': 'CHOICE', 'value': 'foo'},
                         'foo': {'type': 'INTEGER', 'value': 1},
                         'ia5string': {'type': 'IA5String', 'value': '{'},
                         'ia5string2': {'type': 'IA5String', 'value': '{'}}}}


"""Unaligned Packed Encoding Rules (UPER) Rust source code codec generator.

"""

from .utils import ENCODER_AND_DECODER_STRUCTS
from .utils import ENCODER_ABORT
from .utils import DECODER_ABORT
from .utils import Generator
from .utils import is_user_type
from .utils import indent_lines
from .utils import dedent_lines
from .utils import make_camel_case
from ...codecs import uper


ENCODER_INIT = '''\
    fn new(dst: &'a mut [u8]) -> Encoder {
        Encoder {
            size: 8 * dst.len(),
            buf: dst,
            pos: 0,
            error: None
        }
    }\
'''

ENCODER_GET_RESULT = '''
    fn get_result(&self) -> Result<usize, Error> {
        if self.error.is_none() {
            return Ok((self.pos + 7) / 8);
        } else {
            return Err(self.error.unwrap());
        }
    }\
'''

ENCODER_ALLOC = '''
    fn alloc(&mut self, size: usize) -> Result<usize, ()> {
        if self.pos + size <= self.size {
            let pos = self.pos;
            self.pos += size;
            Ok(pos)
        } else {
            self.abort(Error::OutOfMemory);
            Err(())
        }
    }\
'''

ENCODER_APPEND_BIT = '''
    fn append_bit(&mut self, value: u8) {
        if let Ok(pos) = self.alloc(1) {
            if pos % 8 == 0 {
                self.buf[pos / 8] = 0;
            }

            self.buf[pos / 8] |= value << (7 - (pos % 8));
        }
    }\
'''

ENCODER_APPEND_BYTES = '''
    fn append_bytes(&mut self, buf: &[u8]) {
        if let Ok(pos) = self.alloc(8 * buf.len()) {
            let byte_pos = pos / 8;
            let pos_in_byte = pos % 8;

            if pos_in_byte == 0 {
                self.buf.get_mut(byte_pos..byte_pos + buf.len())
                    .unwrap()
                    .copy_from_slice(buf.get(0..buf.len()).unwrap());
            } else {
                for i in 0..buf.len() {
                    self.buf[byte_pos + i] |= buf[i] >> pos_in_byte;
                    self.buf[byte_pos + i + 1] = buf[i] << (8 - pos_in_byte);
                }
            }
        }
    }\
'''

ENCODER_APPEND_U8 = '''
    fn append_u8(&mut self, value: u8) {
        self.append_bytes(&value.to_be_bytes());
    }\
'''

ENCODER_APPEND_U16 = '''
    fn append_u16(&mut self, value: u16) {
        self.append_bytes(&value.to_be_bytes());
    }\
'''

ENCODER_APPEND_U32 = '''
    fn append_u32(&mut self, value: u32) {
        self.append_bytes(&value.to_be_bytes());
    }\
'''

ENCODER_APPEND_U64 = '''
    fn append_u64(&mut self, value: u64) {
        self.append_bytes(&value.to_be_bytes());
    }\
'''

ENCODER_APPEND_I8 = '''
    fn append_i8(&mut self, value: i8) {
        self.append_u8((value as u8).wrapping_add(128));
    }\
'''

ENCODER_APPEND_I16 = '''
    fn append_i16(&mut self, value: i16) {
        self.append_u16((value as u16).wrapping_add(32768));
    }\
'''

ENCODER_APPEND_I32 = '''
    fn append_i32(&mut self, value: i32) {
        self.append_u32((value as u32).wrapping_add(2147483648));
    }\
'''

ENCODER_APPEND_I64 = '''
    fn append_i64(&mut self, value: i64) {
        self.append_u64((value as u64).wrapping_add(9223372036854775808));
    }\
'''

ENCODER_APPEND_BOOL = '''
    fn append_bool(&mut self, value: bool) {
        self.append_bit(value as u8);
    }\
'''

ENCODER_APPEND_NON_NEGATIVE_BINARY_INTEGER = '''
    fn append_non_negative_binary_integer(&mut self, value: u64, size: usize) {
        for i in 0..size {
            self.append_bit((value >> (size - i - 1)) as u8 & 1);
        }
    }\
'''

DECODER_INIT = '''
    fn new(src: &'a[u8]) -> Decoder {
        Decoder {
            buf: src,
            size: 8 * src.len(),
            pos: 0,
            error: None
        }
    }\
'''

DECODER_GET_RESULT = '''
    fn get_result(&self) -> Result<usize, Error> {
        if self.error.is_none() {
            Ok((self.pos + 7) / 8)
        } else {
            Err(self.error.unwrap())
        }
    }\
'''

DECODER_FREE = '''
    fn free(&mut self, size: usize) -> Result<usize, ()> {
        if self.pos + size <= self.size {
            let pos = self.pos;
            self.pos += size;
            Ok(pos)
        } else {
            self.abort(Error::OutOfData);
            Err(())
        }
    }\
'''

DECODER_READ_BIT = '''
    fn read_bit(&mut self) -> u8 {
        if let Ok(pos) = self.free(1) {
            (self.buf[pos / 8] >> (7 - (pos % 8))) & 1
        } else {
            0
        }
    }\
'''

DECODER_READ_BYTES = '''
    fn read_bytes(&mut self, buf: &mut [u8]) {
        if let Ok(pos) = self.free(8 * buf.len()) {
            let byte_pos = pos / 8;
            let pos_in_byte = pos % 8;

            if pos_in_byte == 0 {
                buf.copy_from_slice(
                    self.buf.get(byte_pos..byte_pos + buf.len()).unwrap());
            } else {
                for i in 0..buf.len() {
                    buf[i] = self.buf[byte_pos + i] << pos_in_byte;
                    buf[i] |= self.buf[byte_pos + i + 1] >> (8 - pos_in_byte);
                }
            }
        }
    }\
'''

DECODER_READ_U8 = '''
    fn read_u8(&mut self) -> u8 {
        let mut buf = [0; 1];

        self.read_bytes(&mut buf);

        u8::from_be_bytes(buf)
    }\
'''

DECODER_READ_U16 = '''
    fn read_u16(&mut self) -> u16 {
        let mut buf = [0; 2];

        self.read_bytes(&mut buf);

        u16::from_be_bytes(buf)
    }\
'''

DECODER_READ_U32 = '''
    fn read_u32(&mut self) -> u32 {
        let mut buf = [0; 4];

        self.read_bytes(&mut buf);

        u32::from_be_bytes(buf)
    }\
'''

DECODER_READ_U64 = '''
    fn read_u64(&mut self) -> u64 {
        let mut buf = [0; 8];

        self.read_bytes(&mut buf);

        u64::from_be_bytes(buf)
    }\
'''

DECODER_READ_I8 = '''
    fn read_i8(&mut self) -> i8 {
        self.read_u8().wrapping_sub(128) as i8
    }\
'''

DECODER_READ_I16 = '''
    fn read_i16(&mut self) -> i16 {
        self.read_u16().wrapping_sub(32768) as i16
    }\
'''

DECODER_READ_I32 = '''
    fn read_i32(&mut self) -> i32 {
        self.read_u32().wrapping_sub(2147483648) as i32
    }\
'''

DECODER_READ_I64 = '''
    fn read_i64(&mut self) -> i64 {
        self.read_u64().wrapping_sub(9223372036854775808) as i64
    }\
'''

DECODER_READ_BOOL = '''
    fn read_bool(&mut self) -> bool {
        self.read_bit() != 0
    }\
'''

DECODER_READ_NON_NEGATIVE_BINARY_INTEGER = '''
    fn read_non_negative_binary_integer(&mut self, size: usize) -> u64 {
        let mut value: u64 = 0;

        for _ in 0..size {
            value <<= 1;
            value |= self.read_bit() as u64;
        }

        value
    }\
'''


def does_bits_match_range(number_of_bits, minimum, maximum):
    return 2 ** number_of_bits == (maximum - minimum + 1)


class _Generator(Generator):

    def format_real(self):
        return []

    def get_enumerated_values(self, type_):
        return sorted(type_.root_data_to_index)

    def get_choice_members(self, type_):
        return type_.root_index_to_member.values()

    def format_type(self, type_, checker):
        if isinstance(type_, uper.Integer):
            return self.format_integer(checker)
        elif isinstance(type_, uper.Boolean):
            return self.format_boolean()
        elif isinstance(type_, uper.Real):
            return self.format_real()
        elif isinstance(type_, uper.Null):
            return []
        elif is_user_type(type_):
            return self.format_user_type(type_.type_name,
                                         type_.module_name)
        elif isinstance(type_, uper.OctetString):
            return self.format_octet_string(checker)
        elif isinstance(type_, uper.Sequence):
            return self.format_sequence(type_, checker)
        elif isinstance(type_, uper.Choice):
            return self.format_choice(type_, checker)
        elif isinstance(type_, uper.SequenceOf):
            return self.format_sequence_of(type_, checker)
        elif isinstance(type_, uper.Enumerated):
            return self.format_enumerated(type_)
        else:
            raise self.error(
                "Unsupported type '{}'.".format(type_.type_name))

    def generate_type_declaration_process(self, type_, checker):
        if isinstance(type_, uper.Integer):
            lines = self.format_integer(checker)
            lines[0] += ' value;'
        elif isinstance(type_, uper.Boolean):
            lines = self.format_boolean()
            lines[0] += ' value;'
        elif isinstance(type_, uper.Real):
            lines = self.format_real()
        elif isinstance(type_, uper.Enumerated):
            lines = self.format_enumerated(type_)
            lines[0] += ' value;'
        elif isinstance(type_, uper.Sequence):
            lines = self.format_sequence(type_, checker)
        elif isinstance(type_, uper.SequenceOf):
            lines = self.format_sequence_of(type_, checker)[1:-1]
            lines = dedent_lines(lines)
        elif isinstance(type_, uper.Choice):
            lines = self.format_choice(type_, checker)
        elif isinstance(type_, uper.OctetString):
            lines = self.format_octet_string(checker)[1:-1]
            lines = dedent_lines(lines)
        elif isinstance(type_, uper.Null):
            lines = []
        else:
            raise self.error(
                "Unsupported type '{}'.".format(type_.type_name))

        return lines

    def generate_definition_inner_process(self, type_, checker):
        if isinstance(type_, uper.Integer):
            return self.format_integer_inner(type_, checker)
        elif isinstance(type_, uper.Boolean):
            return self.format_boolean_inner()
        elif isinstance(type_, uper.Real):
            return self.format_real_inner()
        elif isinstance(type_, uper.Sequence):
            return self.format_sequence_inner(type_, checker)
        elif isinstance(type_, uper.SequenceOf):
            return self.format_sequence_of_inner(type_, checker)
        elif isinstance(type_, uper.Choice):
            return self.format_choice_inner(type_, checker)
        elif isinstance(type_, uper.OctetString):
            return self.format_octet_string_inner(type_, checker)
        elif isinstance(type_, uper.Enumerated):
            return self.format_enumerated_inner(type_)
        elif isinstance(type_, uper.Null):
            return self.format_null_inner()
        else:
            return [], []

    def format_integer_inner(self, type_, checker):
        type_name = self.format_type_name(checker.minimum, checker.maximum)
        location = self.location_inner()

        if (type_.number_of_bits % 8) == 0:
            return (
                [
                    'encoder.append_{}(self.{});'.format(
                        type_name,
                        location)
                ],
                [
                    'self.{} = decoder.read_{}();'.format(location, type_name)
                ]
            )
        else:
            return (
                [
                    'encoder.append_non_negative_binary_integer(',
                    '    (self.{} - {}) as u64,'.format(location, checker.minimum),
                    '    {});'.format(type_.number_of_bits)
                ],
                [
                    'self.{} = '
                    'decoder.read_non_negative_binary_integer({}) as {};'.format(
                        location,
                        type_.number_of_bits,
                        type_name),
                    'self.{} += {};'.format(location, checker.minimum)
                ]
            )

    def format_boolean_inner(self):
        return (
            [
                'encoder.append_bool(self.{});'.format(
                    self.location_inner())
            ],
            [
                'self.{} = decoder.read_bool();'.format(
                    self.location_inner())
            ]
        )

    def format_real_inner(self):
        return [], []

    def format_sequence_inner(self, type_, checker):
        encode_lines = []
        decode_lines = []
        member_name_to_is_present = {}

        for member in type_.root_members:
            if member.optional:
                name = '{}is_{}_present'.format(self.location_inner('', '.'),
                                                member.name)
                encode_lines.append(
                    'encoder.append_bit(self.{} as u8);'.format(name))
                decode_lines.append(
                    'self.{} = decoder.read_bit() != 0;'.format(
                        name))
            elif member.default is not None:
                unique_is_present = self.add_unique_variable('is_present')
                member_name_to_is_present[member.name] = unique_is_present
                encode_lines.append(
                    'encoder.append_bit((self.{}{} != {}) as u8);'.format(
                        self.location_inner('', '.'),
                        member.name,
                        member.default))
                decode_lines.append(
                    '{} = decoder.read_bit() == 1;'.format(
                        unique_is_present))

        for member in type_.root_members:
            (member_encode_lines,
             member_decode_lines) = self.format_sequence_inner_member(
                 member,
                 checker,
                 member_name_to_is_present)

            encode_lines += member_encode_lines
            decode_lines += member_decode_lines

        return encode_lines, decode_lines

    def format_octet_string_inner(self, type_, checker):
        location = self.location_inner('', '.')

        if checker.minimum == checker.maximum:
            encode_lines = [
                'encoder.append_bytes(&self.{}buf);'.format(location)
            ]
            decode_lines = [
                'decoder.read_bytes(&mut self.{}buf);'.format(location)
            ]
        else:
            encode_lines = [
                'encoder.append_non_negative_binary_integer(',
                '    (self.{}length - {}) as u64,'.format(location,
                                                          checker.minimum),
                '    {});'.format(type_.number_of_bits),
                'encoder.append_bytes(&self.{}buf,'.format(location),
                '                     self.{}length);'.format(location)
            ]
            decode_lines = [
                'self.{}length = decoder.read_non_negative_binary_integer('.format(
                    location),
                '    {});'.format(type_.number_of_bits),
                'self.{}length += {};'.format(location, checker.minimum)
            ]

            if not does_bits_match_range(type_.number_of_bits,
                                         checker.minimum,
                                         checker.maximum):
                decode_lines += [
                    '',
                    'if self.{}length > {} {{'.format(location, checker.maximum),
                    '    decoder.abort(Error::BadLength);',
                    '',
                    '    return;',
                    '}',
                    ''
                ]

            decode_lines += [
                'decoder.read_bytes(&mut self.{}buf,'.format(location),
                '                   self.{}length);'.format(location)
            ]

        return encode_lines, decode_lines

    def format_user_type_inner(self):
        encode_lines = [
            'self.{}.encode_inner(&mut encoder);'.format(
                self.location_inner())
        ]
        decode_lines = [
            'self.{}.decode_inner(&mut decoder);'.format(
                self.location_inner())
        ]

        return encode_lines, decode_lines

    def format_choice_inner(self, type_, checker):
        encode_lines = []
        decode_lines = []
        unique_choice = self.add_unique_variable('choice')
        choice = '{}choice'.format(self.location_inner('', '.'))

        for member in type_.root_index_to_member.values():
            member_checker = self.get_member_checker(checker,
                                                     member.name)

            with self.asn1_members_backtrace_push(member.name):
                with self.c_members_backtrace_push(member.name):
                    choice_encode_lines, choice_decode_lines = self.format_type_inner(
                        member,
                        member_checker)

            index = type_.root_name_to_index[member.name]

            choice_encode_lines = [
                'encoder.append_non_negative_binary_integer({}, {});'.format(
                    index,
                    type_.root_number_of_bits)
            ] + choice_encode_lines
            encode_lines += [
                '    {}({}) => {{'.format(make_camel_case(member.name),
                                          unique_choice)
            ] + indent_lines(choice_encode_lines, 8) + [
                '    },'
            ]

            choice_decode_lines = [
                'self.{} = {}{};'.format(choice,
                                         self.location,
                                         make_camel_case(member.name))
            ] + choice_decode_lines
            decode_lines += [
                '    {} => {{'.format(index)
            ] + indent_lines(choice_decode_lines, 8) + [
                '    },'
            ]

        encode_lines = [
            '',
            'match self {'
        ] + encode_lines + [
            '    _ => encoder.abort(Error::BadChoice);',
            '}',
            ''
        ]

        decode_lines = [
            '',
            'match decoder.read_non_negative_binary_integer({}) {{'.format(
                type_.root_number_of_bits)
        ] + decode_lines + [
            '    _ => decoder.abort(Error::BadChoice);',
            '}',
            ''
        ]

        return encode_lines, decode_lines

    def format_enumerated_inner(self, type_):
        return (
            [
                'encoder.append_non_negative_binary_integer('
                'self.{}, {});'.format(self.location_inner(),
                                       type_.root_number_of_bits)
            ],
            [
                'self.{} = decoder.read_non_negative_binary_integer('
                '{});'.format(self.location_inner(),
                              type_.root_number_of_bits)
            ]
        )

    def format_null_inner(self):
        return (
            [
                '(void)src_p;'
            ],
            [
                '(void)dst_p;'
            ]
        )

    def format_sequence_of_inner(self, type_, checker):
        unique_i = self.add_unique_variable('i')

        with self.c_members_backtrace_push('elements[{}]'.format(unique_i)):
            encode_lines, decode_lines = self.format_type_inner(
                type_.element_type,
                checker.element_type)

        if checker.minimum == checker.maximum:
            first_encode_lines = first_decode_lines = [
                '',
                'for {} in 0..{} {{'.format(unique_i, checker.maximum)
            ]
        else:
            location = self.location_inner('', '.')
            first_encode_lines = [
                'encoder.append_non_negative_binary_integer(',
                '    (self.{}length - {}) as u64,'.format(location,
                                                          checker.minimum),
                '    {});'.format(type_.number_of_bits),
                '',
                'for {} in 0..self.{}length {{'.format(unique_i, location)
            ]
            first_decode_lines = [
                'self.{}length = '
                'decoder.read_non_negative_binary_integer({});'.format(
                    location,
                    type_.number_of_bits),
                'self.{}length += {};'.format(location, checker.minimum),
                ''
            ]

            if not does_bits_match_range(type_.number_of_bits,
                                         checker.minimum,
                                         checker.maximum):
                first_decode_lines += [
                    'if self.{}length > {} {{'.format(location, checker.maximum),
                    '    decoder.abort(Error::BadLength);',
                    '',
                    '    return;',
                    '}',
                    ''
                ]

            first_decode_lines += [
                'for {} in 0..self.{}length {{'.format(unique_i, location)
            ]

        encode_lines = first_encode_lines + indent_lines(encode_lines) + ['}', '']
        decode_lines = first_decode_lines + indent_lines(decode_lines) + ['}', '']

        return encode_lines, decode_lines

    def format_type_inner(self, type_, checker):
        if isinstance(type_, uper.Integer):
            return self.format_integer_inner(type_, checker)
        elif isinstance(type_, uper.Real):
            return self.format_real_inner()
        elif isinstance(type_, uper.Null):
            return [], []
        elif isinstance(type_, uper.Boolean):
            return self.format_boolean_inner()
        elif is_user_type(type_):
            return self.format_user_type_inner()
        elif isinstance(type_, uper.OctetString):
            return self.format_octet_string_inner(type_, checker)
        elif isinstance(type_, uper.Sequence):
            return self.format_sequence_inner(type_, checker)
        elif isinstance(type_, uper.Choice):
            return self.format_choice_inner(type_, checker)
        elif isinstance(type_, uper.SequenceOf):
            return self.format_sequence_of_inner(type_, checker)
        elif isinstance(type_, uper.Enumerated):
            return self.format_enumerated_inner(type_)
        else:
            raise self.error(type_)

    def generate_encoder(self, definitions):
        functions = [
            ('encoder.init(', ENCODER_INIT),
            ('encoder.get_result(', ENCODER_GET_RESULT),
            ('encoder.abort(', ENCODER_ABORT),
            ('encoder.append_bit(', ENCODER_ALLOC),
            ('encoder.append_bit(', ENCODER_APPEND_BIT),
            ('encoder.append_bytes(', ENCODER_APPEND_BYTES),
            ('encoder.append_u8(', ENCODER_APPEND_U8),
            ('encoder.append_u16(', ENCODER_APPEND_U16),
            ('encoder.append_u32(', ENCODER_APPEND_U32),
            ('encoder.append_u64(', ENCODER_APPEND_U64),
            ('encoder.append_i8(', ENCODER_APPEND_I8),
            ('encoder.append_i16(', ENCODER_APPEND_I16),
            ('encoder.append_i32(', ENCODER_APPEND_I32),
            ('encoder.append_i64(', ENCODER_APPEND_I64),
            ('encoder.append_bool(', ENCODER_APPEND_BOOL),
            (
                'encoder.append_non_negative_binary_integer(',
                ENCODER_APPEND_NON_NEGATIVE_BINARY_INTEGER
            )
        ]

        helpers = [
            "impl<'a> Encoder<'a> {",
            "    fn new(dst: &'a mut [u8]) -> Encoder {",
            '        Encoder {',
            '            size: 8 * dst.len(),',
            '            buf: dst,',
            '            pos: 0,',
            '            error: None',
            '        }',
            '    }'
        ]

        for pattern, definition in functions:
            if pattern in definitions:
                helpers.append(definition)

        return helpers + ['}', '']

    def generate_decoder(self, definitions):
        functions = [
            ('decoder.init(', DECODER_INIT),
            ('decoder.get_result(', DECODER_GET_RESULT),
            ('decoder.abort(', DECODER_ABORT),
            ('decoder.read_bit(', DECODER_FREE),
            ('decoder.read_bit(', DECODER_READ_BIT),
            ('decoder.read_bytes(', DECODER_READ_BYTES),
            ('decoder.read_u8(', DECODER_READ_U8),
            ('decoder.read_u16(', DECODER_READ_U16),
            ('decoder.read_u32(', DECODER_READ_U32),
            ('decoder.read_u64(', DECODER_READ_U64),
            ('decoder.read_i8(', DECODER_READ_I8),
            ('decoder.read_i16(', DECODER_READ_I16),
            ('decoder.read_i32(', DECODER_READ_I32),
            ('decoder.read_i64(', DECODER_READ_I64),
            ('decoder.read_bool(', DECODER_READ_BOOL),
            (
                'decoder.read_non_negative_binary_integer(',
                DECODER_READ_NON_NEGATIVE_BINARY_INTEGER
            )
        ]

        helpers = [
            "impl<'a> Decoder<'a> {",
            "    fn new(src: &'a[u8]) -> Decoder {",
            '        Decoder {',
            '            buf: src,',
            '            size: 8 * src.len(),',
            '            pos: 0,',
            '            error: None',
            '        }',
            '    }'
        ]

        for pattern, definition in functions:
            if pattern in definitions:
                helpers.append(definition)

        return helpers + ['}']

    def generate_helpers(self, definitions):
        helpers = [ENCODER_AND_DECODER_STRUCTS]
        helpers += self.generate_encoder(definitions)
        helpers += self.generate_decoder(definitions)

        return helpers + ['']


def generate(compiled):
    return _Generator().generate(compiled)



"""The top level of the ASN.1 tools package contains commonly used
functions and classes, and the command line interface.

"""

import sys
import os
import argparse
import binascii
import logging
from pprint import pformat

from prompt_toolkit.completion import WordCompleter
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from .compiler import compile_dict
from .compiler import compile_string
from .compiler import compile_files
from .compiler import pre_process_dict
from .parser import parse_string
from .parser import parse_files
from .parser import ParseError
from .errors import Error
from .errors import EncodeError
from .errors import DecodeError
from .errors import CompileError
from .errors import ConstraintsError
from . source import c
from . source import rust
from .version import __version__


__author__ = 'Erik Moqvist'


class ArgumentParserError(Error):
    pass


class ArgumentParser(argparse.ArgumentParser):

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stdout)

        raise ArgumentParserError(message)

    def error(self, message):
        self.print_usage(sys.stdout)
        message = '{}: error: {}'.format(self.prog, message)
        print(message)

        raise ArgumentParserError(message)


def _import_module(pyfilepath):
    module_name = os.path.splitext(os.path.basename(pyfilepath))[0]

    if sys.version_info > (3, 4):
        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, pyfilepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        import imp
        module = imp.load_source(module_name, pyfilepath)

    return module


def _convert_hexstring(input_spec,
                       output_spec,
                       output_codec,
                       type_name,
                       hexstring):
    try:
        encoded = binascii.unhexlify(hexstring)
    except Exception as e:
        raise TypeError("'{}': {}".format(hexstring, str(e)))

    decoded = input_spec.decode(type_name, encoded)

    if output_codec in ['gser', 'xer', 'jer']:
        decoded = output_spec.encode(type_name, decoded, indent=4).strip()
    else:
        decoded = binascii.hexlify(output_spec.encode(type_name, decoded))

    print(decoded.decode('latin-1'))


def _compile_files(specs, input_codec, output_codec, cache_dir):
    py_count = [spec.endswith('.py') for spec in specs].count(True)

    if py_count > 0:
        if py_count != 1:
            raise Exception('Expected one .py-file, but got {}.'.format(py_count))

        module = _import_module(specs[0])
        parsed = module.SPECIFICATION
        input_spec = compile_dict(parsed, input_codec)
        output_spec = compile_dict(parsed, output_codec)
    else:
        input_spec = compile_files(specs,
                                   input_codec,
                                   cache_dir=cache_dir)
        output_spec = compile_files(specs,
                                    output_codec,
                                    cache_dir=cache_dir)

    return input_spec, output_spec


def _do_convert(args):
    input_spec, output_spec = _compile_files(args.specification,
                                             args.input_codec,
                                             args.output_codec,
                                             args.cache_dir)

    if args.hexstring == '-':
        for hexstring in sys.stdin:
            hexstring = hexstring.strip('\r\n')

            if hexstring:
                try:
                    _convert_hexstring(input_spec,
                                       output_spec,
                                       args.output_codec,
                                       args.type,
                                       hexstring)
                except TypeError:
                    print(hexstring)
                except DecodeError as e:
                    print(hexstring)
                    print(str(e))
            else:
                print(hexstring)
    else:
        _convert_hexstring(input_spec,
                           output_spec,
                           args.output_codec,
                           args.type,
                           args.hexstring)


def _handle_command_compile(line):
    parser = ArgumentParser(prog='compile')
    parser.add_argument('-i', '--input-codec',
                        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer'),
                        default='ber',
                        help='Input codec (default: %(default)s).')
    parser.add_argument('-o', '--output-codec',
                        choices=(
                            'ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer', 'gser'
                        ),
                        default='gser',
                        help='Output codec (default: %(default)s).')
    parser.add_argument('-c', '--cache-dir',
                        help='Cache directory.')
    parser.add_argument('specification',
                        nargs='+',
                        help='ASN.1 specification as one or more .asn files.')

    try:
        args = parser.parse_args(args=line.split()[1:])
    except ArgumentParserError:
        return None, None, None

    try:
        input_spec, output_spec = _compile_files(args.specification,
                                                 args.input_codec,
                                                 args.output_codec,
                                                 args.cache_dir)
        return input_spec, output_spec, args.output_codec
    except Exception as e:
        print('error: {}'.format(str(e)))

        return None, None, None


def _handle_command_convert(line, input_spec, output_spec, output_codec):
    if not input_spec:
        print("No compiled specification found. Please use the "
              "'compile' command to compile one.")
        return

    parser = ArgumentParser(prog='convert')
    parser.add_argument('type', help='Type to convert.')
    parser.add_argument(
        'hexstring',
        help='Hexstring to convert.')

    try:
        args = parser.parse_args(args=line.split()[1:])
    except ArgumentParserError:
        return

    try:
        _convert_hexstring(input_spec,
                           output_spec,
                           output_codec,
                           args.type,
                           args.hexstring)
    except Exception as e:
        print('error: {}'.format(str(e)))


def _handle_command_help():
    print('Commands:')
    print('  compile')
    print('  convert')
    print('  exit')
    print('  help')


def _do_shell(_args):
    commands = ['compile', 'convert', 'help', 'exit']
    completer = WordCompleter(commands, WORD=True)
    user_home = os.path.expanduser('~')
    history = FileHistory(os.path.join(user_home, '.asn1tools-history.txt'))
    session = PromptSession(completer=completer,
                            complete_while_typing=True,
                            auto_suggest=AutoSuggestFromHistory(),
                            enable_history_search=True,
                            history=history)
    input_spec = None
    output_spec = None
    output_codec = None

    print("\nWelcome to the asn1tools shell!\n")

    while True:
        try:
            line = session.prompt(u'$ ')
        except EOFError:
            return

        line = line.strip()

        if line:
            if line.startswith('compile'):
                input_spec, output_spec, output_codec = _handle_command_compile(line)
            elif line.startswith('convert'):
                _handle_command_convert(line,
                                        input_spec,
                                        output_spec,
                                        output_codec)
            elif line == 'help':
                _handle_command_help()
            elif line == 'exit':
                return
            else:
                print('{}: command not found'.format(line))


def _do_parse(args):
    parsed = parse_files(args.specification)

    with open(args.outfile, 'w') as fout:
        fout.write('SPECIFICATION = {}'.format(pformat(parsed)))


def _do_generate_c_source(args):
    if args.namespace is None:
        name = os.path.basename(args.specification[0])
        name = os.path.splitext(name)[0]
    else:
        name = args.namespace

    filename_h = name + '.h'
    filename_c = name + '.c'
    fuzzer_filename_c = name + '_fuzzer.c'
    fuzzer_filename_mk = name + '_fuzzer.mk'

    compiled = compile_files(args.specification,
                             args.codec)
    header, source, fuzzer_source, fuzzer_makefile = c.generate(
        compiled,
        args.codec,
        name,
        filename_h,
        filename_c,
        fuzzer_filename_c)

    with open(filename_h, 'w') as fout:
        fout.write(header)

    with open(filename_c, 'w') as fout:
        fout.write(source)

    print('Successfully generated {} and {}.'.format(filename_h,
                                                     filename_c))

    if args.generate_fuzzer:
        with open(fuzzer_filename_c, 'w') as fout:
            fout.write(fuzzer_source)

        with open(fuzzer_filename_mk, 'w') as fout:
            fout.write(fuzzer_makefile)

        print('Successfully generated {} and {}.'.format(fuzzer_filename_c,
                                                         fuzzer_filename_mk))
        print()
        print(
            'Run "make -f {}" to build and run the fuzzer. Requires a'.format(
                fuzzer_filename_mk))
        print('recent version of clang.')


def _do_generate_rust_source(args):
    name = os.path.basename(args.specification[0])
    name = os.path.splitext(name)[0]
    filename_rs = name + '.rs'

    compiled = compile_files(args.specification,
                             args.codec)
    source = rust.generate(compiled, args.codec)

    with open(filename_rs, 'w') as fout:
        fout.write(source)

    print('Successfully generated {}.'.format(filename_rs))


def _main():
    parser = argparse.ArgumentParser(
        description='Various ASN.1 utilities.')

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose',
                        choices=(0, 1, 2),
                        type=int,
                        default=1,
                        help=('Control the verbosity; disable(0), warning(1) '
                              'and debug(2) (default: %(default)s).'))
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # The 'convert' subparser.
    subparser = subparsers.add_parser(
        'convert',
        description='Convert given hextring and print it to standard output.')
    subparser.add_argument(
        '-i', '--input-codec',
        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer'),
        default='ber',
        help='Input format (default: %(default)s).')
    subparser.add_argument(
        '-o', '--output-codec',
        choices=('ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer', 'gser'),
        default='gser',
        help='Output format (default: %(default)s).')
    subparser.add_argument('-c', '--cache-dir',
                           help='Cache directory.')
    subparser.add_argument(
        'specification',
        nargs='+',
        help=('ASN.1 specification as one or more .asn files or one .py '
              'file. The .py-file may be created with the parse subcommand.'))
    subparser.add_argument('type', help='Type to convert.')
    subparser.add_argument(
        'hexstring',
        help='Hexstring to convert, or - to read hexstrings from standard input.')
    subparser.set_defaults(func=_do_convert)

    # The 'shell' subparser.
    subparser = subparsers.add_parser('shell',
                                      description='An interactive shell.')
    subparser.set_defaults(func=_do_shell)

    # The 'parse' subparser.
    subparser = subparsers.add_parser('parse',
                                      description='Convert to a Python dictionary.')
    subparser.add_argument('specification',
                           nargs='+',
                           help='ASN.1 specification as one or more .asn files.')
    subparser.add_argument('outfile',
                           help='Output file name.')
    subparser.set_defaults(func=_do_parse)

    # The 'generate_c_source' subparser.
    subparser = subparsers.add_parser(
        'generate_c_source',
        description='Generate C source code from given ASN.1 specification.')
    subparser.add_argument(
        '-c', '--codec',
        choices=('oer', 'uper'),
        default='oer',
        help='Codec to generate code for (default: %(default)s).')
    subparser.add_argument(
        '-n', '--namespace',
        help='Namespace of defines, structs, functions, files, etc.')
    subparser.add_argument(
        '-f', '--generate-fuzzer',
        action='store_true',
        help='Also generate fuzzer source code.')
    subparser.add_argument('specification',
                           nargs='+',
                           help='ASN.1 specification as one or more .asn files.')
    subparser.set_defaults(func=_do_generate_c_source)

    # The 'generate_rust_source' subparser.
    subparser = subparsers.add_parser(
        'generate_rust_source',
        description='Generate Rust source code from given ASN.1 specification.')
    subparser.add_argument(
        '-c', '--codec',
        choices=('uper', ),
        default='uper',
        help='Codec to generate code for (default: %(default)s).')
    subparser.add_argument('specification',
                           nargs='+',
                           help='ASN.1 specification as one or more .asn files.')
    subparser.set_defaults(func=_do_generate_rust_source)

    args = parser.parse_args()

    levels = [logging.CRITICAL, logging.WARNING, logging.DEBUG]
    level = levels[args.verbose]

    logging.basicConfig(format='%(levelname)s: %(message)s',
                        level=level)

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit('error: {}'.format(str(e)))



EXPECTED = {'X683': {'extensibility-implied': False,
          'imports': {},
          'object-classes': {},
          'object-sets': {},
          'types': {'ErrorCodeString': {'size': [4], 'type': 'IA5String'},
                    'IntegerList1': {'actual-parameters': [{'type': 'INTEGER'}],
                                     'type': 'List1'},
                    'List1': {'members': [{'name': 'elem',
                                           'type': 'ElementTypeParam'},
                                          {'actual-parameters': [{'type': 'ElementTypeParam'}],
                                           'name': 'next',
                                           'optional': True,
                                           'type': 'List1'}],
                              'parameters': ['ElementTypeParam'],
                              'type': 'SEQUENCE'},
                    'OPTIONALLY-SIGNED': {'members': [{'name': 'unsigned-data',
                                                       'tag': {'number': 0},
                                                       'type': 'ToBeSigned'},
                                                      {'actual-parameters': [{'type': 'ToBeSigned'}],
                                                       'name': 'signed-data',
                                                       'tag': {'number': 1},
                                                       'type': 'SIGNED'}],
                                          'parameters': ['ToBeSigned'],
                                          'type': 'CHOICE'},
                    'OptionallySignedUTF8String': {'actual-parameters': [{'type': 'UTF8String'}],
                                                   'type': 'OPTIONALLY-SIGNED'},
                    'SIGNED': {'members': [{'name': 'authenticated-data',
                                            'type': 'ToBeSigned'},
                                           {'name': 'authenticator',
                                            'type': 'BIT STRING'}],
                               'parameters': ['ToBeSigned'],
                               'type': 'SEQUENCE'},
                    'SignedUTF8String': {'actual-parameters': [{'type': 'UTF8String'}],
                                         'type': 'SIGNED'}},
          'values': {}}}